@@ -30,6 +30,7 @@
 #include <ctype.h>
 #include "gmime-part.h"
 #include "gmime-utils.h"
+#include "md5-utils.h"
 
 
 /**
@@ -82,8 +83,12 @@ g_mime_part_new_with_type (const gchar *type, const gchar *subtype)
 void
 g_mime_part_destroy (GMimePart *mime_part)
 {
+	g_return_if_fail (mime_part != NULL);
+	
 	g_free (mime_part->description);
 	g_free (mime_part->content_id);
+	g_free (mime_part->content_md5);
+	g_free (mime_part->content_location);
 	
 	if (mime_part->mime_type)
 		g_mime_content_type_destroy (mime_part->mime_type);
@@ -112,8 +117,6 @@ g_mime_part_destroy (GMimePart *mime_part)
 		}
 	}
 	
-	g_free (mime_part->boundary);
-	
 	if (mime_part->children) {
 		GList *child;
 		
@@ -143,6 +146,8 @@ g_mime_part_destroy (GMimePart *mime_part)
 void
 g_mime_part_set_content_description (GMimePart *mime_part, const gchar *description)
 {
+	g_return_if_fail (mime_part != NULL);
+	
 	if (mime_part->description)
 		g_free (mime_part->description);
 	
@@ -157,8 +162,10 @@ g_mime_part_set_content_description (GMimePart *mime_part, const gchar *descript
  * Returns the content description for the specified mime part.
  **/
 const gchar *
-g_mime_part_get_content_description (GMimePart *mime_part)
+g_mime_part_get_content_description (const GMimePart *mime_part)
 {
+	g_return_val_if_fail (mime_part != NULL, NULL);
+	
 	return mime_part->description;
 }
 
@@ -173,6 +180,8 @@ g_mime_part_get_content_description (GMimePart *mime_part)
 void
 g_mime_part_set_content_id (GMimePart *mime_part, const gchar *content_id)
 {
+	g_return_if_fail (mime_part != NULL);
+	
 	if (mime_part->content_id)
 		g_free (mime_part->content_id);
 	
@@ -189,10 +198,129 @@ g_mime_part_set_content_id (GMimePart *mime_part, const gchar *content_id)
 const gchar *
 g_mime_part_get_content_id (GMimePart *mime_part)
 {
+	g_return_val_if_fail (mime_part != NULL, NULL);
+	
 	return mime_part->content_id;
 }
 
 
+/**
+ * g_mime_part_set_content_md5: Set the content md5
+ * @mime_part: Mime part
+ * @content_md5: content md5 or NULL to generate the md5 digest.
+ *
+ * Set the content md5 for the specified mime part.
+ **/
+void
+g_mime_part_set_content_md5 (GMimePart *mime_part, const gchar *content_md5)
+{
+	const GMimeContentType *type;
+	
+	g_return_if_fail (mime_part != NULL);
+	
+	/* RFC 1864 states that you cannot set a Content-MD5 for these types */
+	type = g_mime_part_get_content_type (mime_part);
+	if (g_mime_content_type_is_type (type, "multipart", "*") ||
+	    g_mime_content_type_is_type (type, "message", "rfc822"))
+		return;
+	
+	if (mime_part->content_md5)
+		g_free (mime_part->content_md5);
+	
+	if (content_md5) {
+		mime_part->content_md5 = g_strdup (content_md5);
+	} else if (mime_part->content && mime_part->content->len) {
+		char digest[16], b64digest[32];
+		int len, state, save;
+		
+		md5_get_digest (mime_part->content->data, mime_part->content->len, digest);
+		
+		state = save = 0;
+		len = g_mime_utils_base64_encode_close (digest, 16, b64digest, &state, &save);
+		b64digest[len] = '\0';
+		
+		mime_part->content_md5 = g_strdup (b64digest);
+	}
+}
+
+
+/**
+ * g_mime_part_verify_content_md5: Verify the content md5
+ * @mime_part: Mime part
+ *
+ * Verify the content md5 for the specified mime part.
+ *
+ * Returns TRUE if the md5 is valid or FALSE otherwise. Note: will
+ * return FALSE if the mime part does not contain a Content-MD5.
+ **/
+gboolean
+g_mime_part_verify_content_md5 (GMimePart *mime_part)
+{
+	char digest[16], b64digest[32];
+	int len, state, save;
+	
+	g_return_val_if_fail (mime_part != NULL, FALSE);
+	g_return_val_if_fail (mime_part->content_md5 != NULL, FALSE);
+	
+	md5_get_digest (mime_part->content->data, mime_part->content->len, digest);
+	
+	state = save = 0;
+	len = g_mime_utils_base64_encode_close (digest, 16, b64digest, &state, &save);
+	b64digest[len] = '\0';
+	
+	return !strcmp (b64digest, mime_part->content_md5);
+}
+
+
+/**
+ * g_mime_part_get_content_md5: Get the content md5
+ * @mime_part: Mime part
+ *
+ * Returns the content md5 for the specified mime part.
+ **/
+const gchar *
+g_mime_part_get_content_md5 (GMimePart *mime_part)
+{
+	g_return_val_if_fail (mime_part != NULL, NULL);
+	
+	return mime_part->content_md5;
+}
+
+
+/**
+ * g_mime_part_set_content_location: Set the content location
+ * @mime_part: Mime part
+ * @content_location: content location
+ *
+ * Set the content location for the specified mime part.
+ **/
+void
+g_mime_part_set_content_location (GMimePart *mime_part, const gchar *content_location)
+{
+	g_return_if_fail (mime_part != NULL);
+	
+	if (mime_part->content_location)
+		g_free (mime_part->content_location);
+	
+	mime_part->content_location = g_strdup (content_location);
+}
+
+
+/**
+ * g_mime_part_get_content_location: Get the content location
+ * @mime_part: Mime part
+ *
+ * Returns the content location for the specified mime part.
+ **/
+const gchar *
+g_mime_part_get_content_location (GMimePart *mime_part)
+{
+	g_return_val_if_fail (mime_part != NULL, NULL);
+	
+	return mime_part->content_location;
+}
+
+
 /**
  * g_mime_part_set_content_type: Set the content type/subtype
  * @mime_part: Mime part
@@ -203,6 +331,8 @@ g_mime_part_get_content_id (GMimePart *mime_part)
 void
 g_mime_part_set_content_type (GMimePart *mime_part, GMimeContentType *mime_type)
 {
+	g_return_if_fail (mime_part != NULL);
+	
 	if (mime_part->mime_type)
 		g_mime_content_type_destroy (mime_part->mime_type);
 	
@@ -219,6 +349,8 @@ g_mime_part_set_content_type (GMimePart *mime_part, GMimeContentType *mime_type)
 const GMimeContentType *
 g_mime_part_get_content_type (GMimePart *mime_part)
 {
+	g_return_val_if_fail (mime_part != NULL, NULL);
+	
 	return mime_part->mime_type;
 }
 
@@ -236,6 +368,8 @@ g_mime_part_get_content_type (GMimePart *mime_part)
 void
 g_mime_part_set_encoding (GMimePart *mime_part, GMimePartEncodingType encoding)
 {
+	g_return_if_fail (mime_part != NULL);
+	
 	mime_part->encoding = encoding;
 }
 
@@ -253,6 +387,8 @@ g_mime_part_set_encoding (GMimePart *mime_part, GMimePartEncodingType encoding)
 GMimePartEncodingType
 g_mime_part_get_encoding (GMimePart *mime_part)
 {
+	g_return_val_if_fail (mime_part != NULL, GMIME_PART_ENCODING_DEFAULT);
+	
 	return mime_part->encoding;
 }
 
@@ -320,6 +456,8 @@ g_mime_part_encoding_from_string (const gchar *encoding)
 void
 g_mime_part_set_content_disposition (GMimePart *mime_part, const gchar *disposition)
 {
+	g_return_if_fail (mime_part != NULL);
+	
 	if (mime_part->disposition) {
 		g_free (mime_part->disposition->disposition);
 		mime_part->disposition->disposition = g_strdup (disposition);
@@ -342,6 +480,8 @@ g_mime_part_set_content_disposition (GMimePart *mime_part, const gchar *disposit
 const gchar *
 g_mime_part_get_content_disposition (GMimePart *mime_part)
 {
+	g_return_val_if_fail (mime_part != NULL, NULL);
+	
 	if (mime_part->disposition)
 		return mime_part->disposition->disposition;
 	
@@ -362,6 +502,8 @@ g_mime_part_add_content_disposition_parameter (GMimePart *mime_part, const gchar
 {
 	GMimeParam *param;
 	
+	g_return_if_fail (mime_part != NULL);
+	
 	if (!mime_part->disposition)
 		g_mime_part_set_content_disposition (mime_part, "");
 	
@@ -420,6 +562,8 @@ g_mime_part_set_filename (GMimePart *mime_part, const gchar *filename)
 {
 	GMimeParam *param;
 	
+	g_return_if_fail (mime_part != NULL);
+	
 	if (!mime_part->disposition)
 		g_mime_part_set_content_disposition (mime_part, "");
 	
@@ -503,8 +647,6 @@ g_mime_part_set_boundary (GMimePart *mime_part, const gchar *boundary)
 	
 	g_return_if_fail (mime_part != NULL);
 	
-	g_free (mime_part->boundary);
-	
 	if (!boundary) {
 		/* Generate a fairly random boundary string. */
 		char digest[16], *p;
@@ -521,7 +663,7 @@ g_mime_part_set_boundary (GMimePart *mime_part, const gchar *boundary)
 		boundary = bbuf;
 	}
 	
-	mime_part->boundary = g_strdup (boundary);
+	g_mime_content_type_add_parameter (mime_part->mime_type, "boundary", boundary);
 }
 
 
@@ -536,7 +678,7 @@ g_mime_part_get_boundary (GMimePart *mime_part)
 {
 	g_return_val_if_fail (mime_part != NULL, NULL);
 	
-	return mime_part->boundary;
+	return g_mime_content_type_get_parameter (mime_part->mime_type, "boundary");
 }
 
 
@@ -549,7 +691,7 @@ g_mime_part_get_boundary (GMimePart *mime_part)
  * Sets the content of the Mime Part (only non-multiparts)
  **/
 void
-g_mime_part_set_content (GMimePart *mime_part, const char *content, guint len)
+g_mime_part_set_content (GMimePart *mime_part, const gchar *content, guint len)
 {
 	g_return_if_fail (mime_part != NULL);
 	
@@ -571,7 +713,7 @@ g_mime_part_set_content (GMimePart *mime_part, const char *content, guint len)
  * Sets the encoding type and raw content on the mime part after decoding the content.
  **/
 void
-g_mime_part_set_pre_encoded_content (GMimePart *mime_part, const char *content, guint len, GMimePartEncodingType encoding)
+g_mime_part_set_pre_encoded_content (GMimePart *mime_part, const gchar *content, guint len, GMimePartEncodingType encoding)
 {
 	gchar *raw;
 	gint save = 0, state = 0;
@@ -615,7 +757,7 @@ g_mime_part_set_pre_encoded_content (GMimePart *mime_part, const char *content,
  * and sets %len to the length of the buffer.
  **/
 const gchar *
-g_mime_part_get_content (GMimePart *mime_part, guint *len)
+g_mime_part_get_content (const GMimePart *mime_part, guint *len)
 {
 	g_return_val_if_fail (mime_part != NULL, NULL);
 	g_return_val_if_fail (mime_part->content != NULL, NULL);
@@ -627,20 +769,24 @@ g_mime_part_get_content (GMimePart *mime_part, guint *len)
 
 
 /**
- * g_mime_part_add_child: Add a child mime part to a multipart
+ * g_mime_part_add_subpart: Add a subpart to a multipart
  * @mime_part: Parent Mime part
- * @child: Child Mime part
+ * @subpart: Child Mime part
  *
- * Adds a child mime part to the parent mime part which *must* be a
+ * Adds a subpart to the parent mime part which *must* be a
  * multipart.
  **/
 void
-g_mime_part_add_child (GMimePart *mime_part, GMimePart *child)
+g_mime_part_add_subpart (GMimePart *mime_part, GMimePart *subpart)
 {
-	g_return_if_fail (mime_part != NULL);
-	g_return_if_fail (child != NULL);
+	const GMimeContentType *type;
 	
-	mime_part->children = g_list_append (mime_part->children, child);
+	g_return_if_fail (mime_part != NULL);
+	g_return_if_fail (subpart != NULL);
+	
+	type = g_mime_part_get_content_type (mime_part);
+	if (g_mime_content_type_is_type (type, "multipart", "*"))
+		mime_part->children = g_list_append (mime_part->children, subpart);
 }
 
 static gchar *
@@ -724,7 +870,10 @@ get_content_type (GMimeContentType *mime_type)
 	str = string->str;
 	g_string_free (string, FALSE);
 	
-	return str;
+	type = g_mime_utils_header_printf ("Content-Type: %s\n", str);
+	g_free (str);
+	
+	return type;
 }
 
 static gchar *
@@ -766,21 +915,26 @@ get_content (GMimePart *part)
  * headers for rfc822 messages.
  **/
 gchar *
-g_mime_part_to_string (GMimePart *mime_part, gboolean toplevel) {
-	gchar *string;
+g_mime_part_to_string (GMimePart *mime_part, gboolean toplevel)
+{
+	gchar *str;
 	
 	g_return_val_if_fail (mime_part != NULL, NULL);
 	
 	if (g_mime_content_type_is_type (mime_part->mime_type, "multipart", "*")) {
+		const gchar *boundary;
 		GString *contents;
 		GList *child;
 		gchar *content_type;
 		
 		/* make sure there's a boundary, else force a random boundary */
-		if (!mime_part->boundary)
+		boundary = g_mime_part_get_boundary (mime_part);
+		if (!boundary) {
 			g_mime_part_set_boundary (mime_part, NULL);
+			boundary = g_mime_part_get_boundary (mime_part);
+		}
 		
-		content_type = g_mime_content_type_to_string (mime_part->mime_type);
+		content_type = get_content_type (mime_part->mime_type);
 		
 		contents = g_string_new ("");
 		
@@ -790,7 +944,7 @@ g_mime_part_to_string (GMimePart *mime_part, gboolean toplevel) {
 			
 			mime_string = g_mime_part_to_string (child->data, FALSE);
 			
-			part = g_strdup_printf ("--%s\n%s\n", mime_part->boundary, mime_string);
+			part = g_strdup_printf ("--%s\n%s\n", boundary, mime_string);
 			
 			g_string_append (contents, part);
 			
@@ -802,17 +956,19 @@ g_mime_part_to_string (GMimePart *mime_part, gboolean toplevel) {
 		
 		if (toplevel) {
 			/* add message header stuff first */
-			string = g_strdup_printf ("MIME-Version: 1.0\n"
-						  "Content-Type: %s; boundary=\"%s\"\n\n"
-						  "This is a multi-part message in MIME format.\n\n"
-						  "%s\n--%s--\n",
-						  content_type, mime_part->boundary,
-						  contents->str, mime_part->boundary);
+			str = g_strdup_printf ("MIME-Version: 1.0\n"
+					       "%s\n"  /* content-type */
+					       "This is a multi-part message in MIME format.\n\n"
+					       "%s\n--%s--\n",
+					       content_type,
+					       contents->str,
+					       boundary);
 		} else {
-			string = g_strdup_printf ("Content-Type: %s; boundary=\"%s\"\n\n"
-						  "%s\n--%s--\n",
-						  content_type, mime_part->boundary,
-						  contents->str, mime_part->boundary);
+			str = g_strdup_printf ("%s\n"  /* content-type */
+					       "%s\n--%s--\n",
+					       content_type,
+					       contents->str,
+					       boundary);
 		}
 		
 		g_free (content_type);
@@ -822,67 +978,88 @@ g_mime_part_to_string (GMimePart *mime_part, gboolean toplevel) {
 		gchar *disposition;
 		gchar *description;
 		gchar *content_id;
+		gchar *content_md5;
+		gchar *content_location;
 		gchar *content;
-		gchar *extras;
 		gchar *text;
+		GString *string;
 		
-		content_type = get_content_type (mime_part->mime_type);
-		disposition = get_content_disposition (mime_part);
-		
-		if (mime_part->description) {
-			text = g_mime_utils_8bit_header_encode (mime_part->description);
-			description = g_strdup_printf ("Content-Description: %s\n", text);
-			g_free (text);
-		} else {
-			description = g_strdup ("");
-		}
-		
-		if (mime_part->content_id)
-			content_id = g_strdup_printf ("Content-Id: %s\n", mime_part->content_id);
-		else
-		        content_id = g_strdup ("");
+		string = g_string_new ("");
 		
 		if (toplevel)
-			extras = "MIME-Version: 1.0\n";
-		else
-			extras = "";
+			g_string_append (string, "MIME-Version: 1.0\n");
+		
+		/* Content-Type: */
+		content_type = get_content_type (mime_part->mime_type);
+		g_string_append (string, content_type);
+		g_free (content_type);
+		
+		/* Content-Transfer-Encoding: */
+		if (mime_part->encoding != GMIME_PART_ENCODING_DEFAULT) {
+			gchar *content_encoding;
+			
+			content_encoding = g_strdup_printf ("Content-Transfer-Encoding: %s\n",
+							    g_mime_part_encoding_to_string (mime_part->encoding));
+			
+			g_string_append (string, content_encoding);
+			g_free (content_encoding);
+		}
+		
+		/* Content-Disposition: */
+		disposition = get_content_disposition (mime_part);
+		if (disposition) {
+			text = g_mime_utils_header_printf ("Content-Disposition: %s\n",
+							   disposition);
+			g_free (disposition);
+			g_string_append (string, text);
+			g_free (text);
+		}
+		
+		/* Content-Description: */
+		if (mime_part->description) {
+			text = g_mime_utils_8bit_header_encode (mime_part->description);
+			description = g_mime_utils_header_printf ("Content-Description: %s\n", text);
+			g_free (text);
+			
+			g_string_append (string, description);
+			g_free (description);
+		}
+		
+		/* Content-Location: */
+		if (mime_part->content_location) {
+			content_md5 = g_strdup_printf ("Content-Location: %s\n", mime_part->content_location);
+			g_string_append (string, content_location);
+			g_free (content_location);
+		}
+		
+		/* Content-Md5: */
+		if (mime_part->content_md5) {
+			content_md5 = g_strdup_printf ("Content-MD5: %s\n", mime_part->content_md5);
+			g_string_append (string, content_md5);
+			g_free (content_md5);
+		}
+		
+		/* Content-Id: */
+		if (mime_part->content_id) {
+			content_id = g_strdup_printf ("Content-Id: %s\n", mime_part->content_id);
+			g_string_append (string, content_id);
+			g_free (content_id);
+		}
+		
+		g_string_append_c (string, '\n');
 		
 		content = get_content (mime_part);
 		
-		if (disposition && *disposition) {
-			string = g_strdup_printf ("%sContent-Type: %s\n"
-						  "Content-Transfer-Encoding: %s\n"
-						  "Content-Disposition: %s\n"
-						  "%s"      /* description */
-						  "%s\n"    /* content id */
-						  "%s\n",   /* content */
-						  extras, content_type,
-						  g_mime_part_encoding_to_string (mime_part->encoding),
-						  disposition,
-						  description,
-						  content_id,
-						  content);
-		} else {
-			string = g_strdup_printf ("%sContent-Type: %s\n"
-						  "Content-Transfer-Encoding: %s\n"
-						  "%s"      /* description */
-						  "%s\n"    /* content id */
-						  "%s\n",   /* content */
-						  extras, content_type,
-						  g_mime_part_encoding_to_string (mime_part->encoding),
-						  description,
-						  content_id,
-						  content);
-		}
+		g_string_append (string, content);
+		g_string_append_c (string, '\n');
 		
-		g_free (content_type);
-		g_free (disposition);
-		g_free (description);
-		g_free (content_id);
 		g_free (content);
+		
+		str = string->str;
+		g_string_free (string, FALSE);
 	}
 	
-	return string;
+	return str;
 }
 
 
@@ -912,3 +1089,46 @@ g_mime_part_foreach (GMimePart *mime_part, GMimePartFunc callback, gpointer data
 		}
 	}
 }
+
+
+/**
+ * g_mime_part_get_subpart_from_content_id: 
+ * @mime_part: the MIME part
+ * @content_id: the content id of the part to look for
+ *
+ * Returns the GMimePart whose content-id matches the search string,
+ * or NULL if a match cannot be found.
+ **/
+const GMimePart *
+g_mime_part_get_subpart_from_content_id (GMimePart *mime_part, const gchar *content_id)
+{
+	GList *child;
+	
+	g_return_val_if_fail (mime_part != NULL, NULL);
+	g_return_val_if_fail (content_id != NULL, NULL);
+	
+	if (mime_part->content_id && !strcmp (mime_part->content_id, content_id))
+		return mime_part;
+	
+	child = mime_part->children;
+	while (child) {
+		const GMimeContentType *type;
+		const GMimePart *part = NULL;
+		GMimePart *subpart;
+		
+		subpart = (GMimePart *) child->data;
+		type = g_mime_part_get_content_type (subpart);
+		
+		if (g_mime_content_type_is_type (type, "multipart", "*"))
+			part = g_mime_part_get_subpart_from_content_id (subpart, content_id);
+		else if (subpart->content_id && !strcmp (subpart->content_id, content_id))
+			part = subpart;
+		
+		if (part)
+			return part;
+		
+		child = child->next;
+	}
+	
+	return NULL;
+}
